import datetime, time

from six.moves import urllib
from six import string_types

from stackexchange.web import WebRequestManager
from stackexchange.core import *
from stackexchange.models import *

##### Site metadata ###
# (originally in the stackauth module; now we need it for Site#info)

class SiteState(Enumeration):
    """Describes the state of a StackExchange site."""
    Normal, OpenBeta, ClosedBeta, LinkedMeta = range(4)

class SiteType(Enumeration):
    '''Describes the type (meta or non-meta) of a StackExchange site.'''
    MainSite, MetaSite = range(2)

class MarkdownExtensions(Enumeration):
    '''Specifies one of the possible extensions to Markdown a site can have enabled.'''
    MathJax, Prettify, Balsamiq, JTab = range(4)

class SiteDefinition(JSONModel):
    """Contains information about a StackExchange site, reported by StackAuth."""
    transfer = ('aliases', 'api_site_parameter', 'audience', 'favicon_url',
        'high_resolution_icon_url', 'icon_url', 'logo_url', 'name', 'open_beta_date',
        'related_sites', 'site_state', 'site_type', 'site_url', 'twitter_account',
        'api_site_parameter',
        ('closed_beta_date', UNIXTimestamp),
        ('open_beta_date', UNIXTimestamp),
        ('launch_date', UNIXTimestamp),
        ('markdown_extensions', ListOf(MarkdownExtensions.from_string)),
        ('site_state', SiteState.from_string),
        ('site_type', SiteType.from_string),
        ('styling', DictObject))

    # To maintain API compatibility only; strictly speaking, we should use api_site_parameter
    # to create new sites, and that's what we do in get_site()
    alias = (('api_endpoint', 'site_url'), ('description', 'audience'))


    def _extend(self, json, stackauth):
        # The usual enumeration heuristics need a bit of help to parse the
        # site state as returned by the API
        fixed_state = re.sub(r'_([a-z])', lambda match: match.group(1).upper(), json.site_state)
        fixed_state = fixed_state[0].upper() + fixed_state[1:]
        self.state = SiteState.from_string(fixed_state)
    
    def get_site(self, *a, **kw):
        return Site(self.api_site_parameter, *a, **kw)

##### Statistics    ###
class Statistics(JSONModel):
    """Stores statistics for a StackExchange site."""
    transfer = ('new_active_users', 'total_users', 'badges_per_minute',
        'total_badges', 'total_votes', 'total_comments', 'answers_per_minute',
        'questions_per_minute', 'total_answers', 'total_accepted',
        'total_unanswered', 'total_questions', 'api_revision')
    alias = (('site_definition', 'site', ModelRef(SiteDefinition)),)


class Site(object):
    """Stores information and provides methods to access data on a StackExchange site. This class is the 'root' of the API - all data is accessed
through here."""

    def __init__(self, domain, app_key = None, cache = 1800, impose_throttling = False, force_http = False):
        self.domain = domain
        self.app_key = app_key
        self.api_version = '2.2'

        self.impose_throttling = impose_throttling
        self.throttle_stop = True
        self.cache_options = {'cache': False} if cache == 0 else {'cache': True, 'cache_age': cache}
        self.force_http = force_http

        self.include_body = False
        self.include_comments = False

        # In API v2.x, we generally don't get given api. at the start of these things, nor are they
        # strictly domains in many cases. We continue to accept api.* names for compatibility.
        domain_components = self.domain.split('.')
        if domain_components[0] == 'api':
            self.root_domain = '.'.join(domain_components[1:])
        else:
            self.root_domain = domain

    URL_Roots = {
        User: 'users/%s',
        Badge: 'badges/%s',
        Answer: 'answers/%s',
        Comment: 'comments/%s',
        Question: 'questions/%s',
    }

    def check_filter(self, kw):
        if 'answers' not in kw:
            kw['answers'] = 'true'
        if self.include_body:
            kw['body'] = 'true'
        if self.include_comments:
            kw['comments'] = 'true'

        # for API v2.x, the comments, body and answers parameters no longer
        # exist; instead, we have to use filters. for now, take the easy way
        # out and just rewrite them in terms of the new filters.
        if 'filter' not in kw:
            filter_name = '_'

            if kw.get('body'):
                filter_name += 'b'
                del kw['body']
            if kw.get('comments'):
                filter_name += 'c'
                del kw['comments']
            if kw.get('answers'):
                filter_name += 'a'
                del kw['answers']

            if filter_name == '_ca':
                # every other compatibility filter name works in the above
                # order except this one...
                kw['filter'] = '_ac'
            elif filter_name != '_':
                kw['filter'] = filter_name

    def _kw_to_str(self, ob):
        try:
            if isinstance(ob, datetime.datetime):
                return str(time.mktime(ob.timetuple()))
            elif isinstance(ob, string_types):
                return ob
            else:
                i = iter(ob)
                return ';'.join(i)
        except TypeError:
            return str(ob).lower()

    def _request(self, to, params):
        url = 'api.stackexchange.com/' + self.api_version + '/' + to
        params['site'] = params.get('site', self.root_domain)

        new_params = {}
        for k, v in params.items():
            if v is None:
                pass
            elif k in ('fromdate', 'todate'):
                # bit of a HACKish workaround for a reported issue; force to an integer
                new_params[k] = str(int(v))
            else:
                new_params[k] = self._kw_to_str(v)

        if self.app_key != None:
            new_params['key'] = self.app_key

        request_properties = dict([(x, getattr(self, x)) for x in ('impose_throttling', 'throttle_stop', 'force_http')])
        request_properties.update(self.cache_options)
        request_mgr = WebRequestManager(**request_properties)

        json, info = request_mgr.json_request(url, new_params)

        if 'quota_remaining' in json and 'quota_max' in json:
            self.rate_limit = (json['quota_remaining'], json['quota_max'])
            self.requests_used = self.rate_limit[1] - self.rate_limit[0]
            self.requests_left = self.rate_limit[0]

        return json

    def _user_prop(self, qs, typ, coll, kw, prop = 'user_id'):
        if prop not in kw:
            raise LookupError('No user ID provided.')
        else:
            tid = self.vectorise(kw[prop], User)
            del kw[prop]

            return self.build('users/%s/%s' % (tid, qs), typ, coll, kw)

    def be_inclusive(self):
        """Include the body and comments of a post, where appropriate, by default."""
        self.include_body, self.include_comments = True, True

    def build(self, url, typ, collection, kw = {}):
        """Builds a StackExchangeResultset object from the given URL and type."""
        if 'body' not in kw:
            kw['body'] = str(self.include_body).lower()
        if 'comments' not in kw:
            kw['comments'] = str(self.include_comments).lower()

        json = self._request(url, kw)
        return JSONMangler.json_to_resultset(self, json, typ, collection, (self, url, typ, collection, kw))

    def build_from_snippet(self, json, typ):
        return StackExchangeResultSet([typ(x, self) for x in json])

    def vectorise(self, lst, or_of_type = None):
        # Ensure we're always dealing with an iterable
        allowed_types = or_of_type
        if allowed_types is not None and not hasattr(allowed_types, '__iter__'):
            allowed_types = (allowed_types, )

        if isinstance(lst, string_types) or type(lst).__name__ == 'bytes':
            return lst
        elif hasattr(lst, '__iter__'):
            return ';'.join([self.vectorise(x, or_of_type) for x in lst])
        elif allowed_types is not None and any([isinstance(lst, typ) for typ in allowed_types]) and hasattr(lst, 'id'):
            return str(lst.id)
        else:
            return str(lst).lower()

    def _get(self, typ, ids, coll, kw):
        root = self.URL_Roots[typ] % self.vectorise(ids)
        return self.build(root, typ, coll, kw)


    def user(self, nid, **kw):
        """Retrieves an object representing the user with the ID `nid`."""
        u, = self.users((nid,), **kw)
        return u

    def users(self, ids = [], **kw):
        """Retrieves a list of the users with the IDs specified in the `ids' parameter."""
        return self._get(User, ids, 'users', kw)
    
    def users_by_name(self, name, **kw):
        kw['filter'] = name
        return self.users(**kw)

    def moderators(self, **kw):
        """Retrieves a list of the moderators on the site."""
        return self.build('users/moderators', User, 'users', kw)

    def moderators_elected(self, **kw):
        """Retrieves a list of the elected moderators on the site."""
        return self.build('users/moderators/elected', User, 'users', kw)

    def answer(self, nid, **kw):
        """Retrieves an object describing the answer with the ID `nid`."""
        a, = self.answers((nid,), **kw)
        return a

    def answers(self, ids = None, **kw):
        """Retrieves a set of the answers with the IDs specified in the 'ids' parameter, or by the
        user_id specified."""
        self.check_filter(kw)
        if ids is None and 'user_id' in kw:
            return self._user_prop('answers', Answer, 'answers', kw)
        elif ids is None:
            return self.build('answers', Answer, 'answers', kw)
        else:
            return self._get(Answer, ids, 'answers', kw)

    def comment(self, nid, **kw):
        """Retrieves an object representing a comment with the ID `nid`."""
        c, = self.comments((nid,), **kw)
        return c

    def comments(self, ids = None, posts = None, **kw):
        """Returns all the comments on the site."""
        if ids is None:
            if posts is None:
                return self.build('comments', Comment, 'comments', kw)
            else:
                url = 'posts/%s/comments' % self.vectorise(posts, (Question, Answer))
                return self.build(url, Comment, 'comments', kw)
        else:
            return self.build('comments/%s' % self.vectorise(ids), Comment, 'comments', kw)

    def question(self, nid, **kw):
        """Retrieves an object representing a question with the ID `nid`. Note that an answer ID can not be specified -
unlike on the actual site, you will receive an error rather than a redirect to the actual question."""
        q, = self.questions((nid,), **kw)
        return q

    questions = property(lambda s: QuestionsQuery(s))

    def recent_questions(self, **kw):
        """Returns the set of the most recent questions on the site, by last activity."""
        if 'answers' not in kw:
            kw['answers'] = 'true'
        return self.build('questions', Question, 'questions', kw)

    def badge_recipients(self, bids, **kw):
        """Returns a set of badges recently awarded on the site, constrained to those with the given IDs, with the 'user' property set describing the user to whom it was awarded."""
        return self.build('badges/%s/recipients' % self.vectorise(bids), Badge, 'badges', kw)

    def all_badges(self, **kw):
        """Returns the set of all the badges which can be awarded on the site, excluding those which are awarded for specific tags."""
        return self.build('badges', Badge, 'badges', kw)

    def badges(self, ids = None, **kw):
        """Returns the badge objectss with the given IDs."""
        if ids == None:
            return self._user_prop('badges', Badge, 'users', kw)
        else:
            return self._get(Badge, ids, 'users', kw)

    def badge(self, nid = None, name = None, **kw):
        """Returns an object representing the badge with the ID 'nid', or with the name passed in as name=."""
        if nid is not None and name is None:
            b, = self.build('badges/%d' % nid, Badge, 'badges', kw)
            return b
        elif nid is None and name is not None:
            # We seem to need to get all badges and find it by name. Sigh.
            kw['inname'] = name
            all_badges = self.build('badges', Badge, 'badges', kw)
            for badge in all_badges:
                if badge.name == name:
                    return badge
            return None
        else:
            raise KeyError('Supply exactly one of the following: a badge ID, a badge name')

    def privileges(self, **kw):
        """Returns all the privileges a user can have on the site."""
        return self.build('privileges', Privilege, 'privileges', kw)

    def all_nontag_badges(self, **kw):
        """Returns the set of all badges which are not tag-based."""
        return self.build('badges/name', Badge, 'badges', kw)

    def all_tag_badges(self, **kw):
        """Returns the set of all the tag-based badges: those which are awarded for performance on a specific tag."""
        return self.build('badges/tags', Badge, 'badges', kw)

    def all_tags(self, **kw):
        '''Returns the set of all tags on the site.'''
        return self.build('tags', Tag, 'tags', kw)

    def info(self, **kw):
        '''Returns statistical information and metadata about the site, such as the total number of questions.
        
Call with site=True to receive a SiteDefinition object representing this site in the site_definition field of the result.'''
        if kw.get('site'):
            # we need to remove site as a query parameter anyway to stop API
            # getting confused
            del kw['site']
            if 'filter' not in kw:
                kw['filter'] = '!9YdnSFfpS'

        return self.build('info', Statistics, 'statistics', kw)[0]

    def stats(self, *a, **kw):
        '''An alias for Site.info().'''
        # this is just an alias to info(), since the method name has changed since
        return self.info(*a, **kw)

    def revision(self, post, guid, **kw):
        real_id = post.id if isinstance(post, Question) or isinstance(post, Answer) else post
        return self.build('revisions/%d/%s' % (real_id, guid), PostRevision, 'revisions', kw)[0]

    def revisions(self, post, **kw):
        return self.build('revisions/' + self.vectorise(post, (Question, Answer)), PostRevision, 'revisions', kw)

    def search(self, **kw):
        return self.build('search', Question, 'questions', kw)

    def search_advanced(self, **kw):
        kw['body'] = None
        kw['comments'] = None
        return self.build('search/advanced', Question, 'questions', kw)

    def similar(self, title, tagged = None, nottagged = None, **kw):
        if 'answers' not in kw:
            kw['answers'] = True
        if tagged is not None:
            kw['tagged'] = self.vectorise(tagged, Tag)
        if nottagged is not None:
            kw['nottagged'] = self.vectorise(nottagged, Tag)

        kw['title'] = title
        return self.build('similar', Question, 'questions', kw)

    def tags(self, **kw):
        return self.build('tags', Tag, 'tags', kw)

    def tag(self, tag, **kw):
        return self.build('tags/%s/info' % tag, Tag, 'tags', kw)[0]

    def tag_wiki(self, tag, **kw):
        return self.build('tags/%s/wikis' % tag, TagWiki, 'tags', kw)

    def tag_related(self, tag, **kw):
        return self.build('tags/%s/related' % tag, Tag, 'tags', kw)

    def tag_synonyms(self, **kw):
        return self.build('tags/synonyms', TagSynonym, 'tag_synonyms', kw)

    def error(self, id, **kw):
        # for some reason, the SE API couldn't possible just ignore site=
        kw['site'] = None
        return self._request('errors/%d' % id, kw)

    def __add__(self, other):
        if isinstance(other, Site):
            return CompositeSite(self, other)
        else:
            raise NotImplemented

class CompositeSite(object):
    def __init__(self, s1, s2):
        self.site_one = s1
        self.site_two = s2

    def __getattr__(self, a):
        if hasattr(self.site_one, a) and hasattr(self.site_two, a) and callable(getattr(self.site_one, a)):
            def handle(*ps, **kws):
                res1 = getattr(self.site_one, a)(*ps, **kws)
                res2 = getattr(self.site_two, a)(*ps, **kws)

                if hasattr(res1, '__iter__') and hasattr(res2, '__iter__'):
                    return res1 + res2
                else:
                    return (res1, res2)

            return handle
        else:
            raise AttributeError(a)

    def __sub__(self, other):
        if other is self.site_one:
            return self.site_two
        elif other is self.site_two:
            return self.site_one
        else:
            raise NotImplemented



import re, sys
from stackexchange.core import *

LaterClass = lambda name: LaterClassIn(name, sys.modules[__name__])

#### Revisions #
class RevisionType(Enumeration):
    SingleUser = 'single_user'
    VoteBased  = 'vote_based'

class PostRevision(JSONModel):
    transfer = ('body', 'comment', 'is_question', 'is_rollback', 'last_body',
        'last_title', 'revision_guid', 'revision_number', 'title',
        'set_community_wiki', 'post_id', 'last_tags', 'tags',
        ('creation_date', UNIXTimestamp),
        ('revision_type', RevisionType.from_string))

    def _extend(self, json, site):
        part = json.user
        self.user = User.partial(lambda self: self.site.user(self.id), site, {
            'id': part['user_id'],
            'user_type': Enumeration.from_string(part['user_type'], UserType),
            'display_name': part['display_name'],
            'reputation': part['reputation'],
            'profile_image': part['profile_image']
        })

    def _get_post(self):
        if self.is_question:
            return self.site.question(self.post_id)
        else:
            return self.site.answer(self.post_id)
    post = property(_get_post)

    # The SE API seems quite inconsistent in this regard; the other methods give a post_type in their JSON
    def _get_post_type(self):
        return PostType.Question if self.is_question else PostType.Answer
    post_type = property(_get_post_type)

    def __repr__(self):
        return '<Revision %d of %s%d>' % (self.revision_number, 'Q' if self.is_question else 'A', self.post_id)

class PostType(Enumeration):
    """Denotes the type of a post: a question or an answer."""
    Question, Answer = 'question', 'answer'

## Timeline ##
class TimelineEventType(Enumeration):
    """Denotes the type of a timeline event."""
    _map = {'askoranswered': 'AskOrAnswered'}

    Comment = 'comment'
    AskOrAnswered = 'askoranswered'
    Badge = 'badge'
    Revision = 'revision'
    Accepted = 'accepted'

class TimelineEvent(JSONModel):
    transfer = ('user_id', 'post_id', 'comment_id', 'action', 'description',
        'detail', 'comment_id',
        ('timeline_type', TimelineEventType.from_string),
        ('post_type', PostType.from_string),
        ('creation_date', UNIXTimestamp))

    _post_related = (TimelineEventType.AskOrAnswered, TimelineEventType.Revision, TimelineEventType.Comment)
    def _get_post(self):
        if self.timeline_type in self._post_related:
            if self.post_type == PostType.Question:
                return self.site.question(self.post_id)
            else:
                return self.site.answer(self.post_id)
        else:
            return None

    def _get_comment(self):
        if self.timeline_type == TimelineEventType.Comment:
            return self.site.comment(self.comment_id)
        else:
            return None

    def _get_badge(self):
        if self.timeline_type == TimelineEventType.Badge:
            return self.site.badge(name = self.description)
        else:
            return None

    post = property(_get_post)
    comment = property(_get_comment)
    badge = property(_get_badge)


##### Content Types ###
class Comment(JSONModel):
    """Describes a comment to a question or answer on a StackExchange site."""
    transfer = ('post_id', 'score', 'edit_count', 'body',
        ('creation_date', UNIXTimestamp), ('post_type', PostType.from_string))

    def _extend(self, json, site):
        self.id = json.comment_id

        if hasattr(json, 'owner'):
            self.owner_id = json.owner['owner_id'] if 'owner_id' in json.owner else json.owner['user_id']
            self.owner = User.partial(lambda self: self.site.user(self.id), site, {
                'id': self.owner_id,
                'user_type': Enumeration.from_string(json.owner['user_type'], UserType),
                'display_name': json.owner['display_name'],
                'reputation': json.owner['reputation'],
                'profile_image': json.owner['profile_image']})
        else:
            self.owner = None

        if hasattr(json, 'reply_to'):
            self.reply_to_user_id = json.reply_to['user_id']
            self.reply_to = User.partial(lambda self: self.site.user(self.id), site, {
                'id': self.reply_to_user_id,
                'user_type': Enumeration.from_string(json.reply_to['user_type'], UserType),
                'display_name': json.reply_to['display_name'],
                'reputation': json.reply_to['reputation'],
                'profile_image': json.reply_to['profile_image']})

    @property
    def post(self):
        if self.post_type == PostType.Question:
            return self.site.question(self.post_id)
        elif self.post_type == PostType.Answer:
            return self.site.answer(self.post_id)
        else:
            return None

    def __unicode__(self):
        return u'Comment ' + str(self.id)
    def __str__(self):
        return str(unicode(self))

class Answer(JSONModel):
    """Describes an answer on a StackExchange site."""

    transfer = ('is_accepted', 'locked_date', 'question_id', 'up_vote_count',
        'down_vote_count', 'view_count', 'score', 'community_owned', 'title',
        'body', 'body_markdown', ('creation_date', UNIXTimestamp),
        ('last_edit_date', UNIXTimestamp),
        ('last_activity_date', UNIXTimestamp),
        ('revisions', LazySequenceField(PostRevision, 'posts/{id}/revisions')))
    alias = (('id', 'answer_id'), ('accepted', 'is_accepted'))

    def _extend(self, json, site):
        if not hasattr(json, '_params_'):
            comment = False
        else:
            comment = ('comment' in json._params_ and json._params_['comment'])

        answer_comments_url = 'answers/%d/comments' % self.id
        self.comments = site.build_from_snippet(json.comments, Comment) if comment else StackExchangeLazySequence(Comment, None, site, answer_comments_url, self._up('comments'), filter = '!-*7AsUyrEan0')

        self._question, self._owner = None, None
        if hasattr(json, 'owner'):
            self.owner_id = json.owner.get('user_id')
            self.owner_info = tuple(json.owner.values())

        if hasattr(self, 'up_vote_count') and hasattr(self, 'down_vote_count'):
            self.votes = (self.up_vote_count, self.down_vote_count)

        self.url = 'http://' + self.site.root_domain + '/questions/' + str(self.question_id) + '/' + str(self.id) + '#' + str(self.id)

    def _get_user(self, id):
        if self._owner is None:
            self._owner = self.site.user(id)
        return self._owner

    def _set_user(self, ob):
        self._owner = ob

    def _get_quest(self, id):
        if self._question is None:
            self._question = self.site.question(id)
        return self._question

    def _set_quest(self, ob):
        self._question = ob

    question = property(_get_quest, _set_quest)
    owner = property(_get_user, _set_user)

    def fetch_callback(self, _, site):
        return site.answer(self.id)

    def __unicode__(self):
        return u'Answer %d' % self.id

    def __str__(self):
        return str(unicode(self))

    def __repr__(self):
        return '<Answer %d @ %x>' % (self.id, id(self))

class IDPartial(ComplexTransform):
    def __init__(self, model_type, fetch_callback):
        self.partial = PartialModelRef(model_type, fetch_callback)

    def __call__(self, key, value, model):
        return self.partial(key, {'id': value}, model)

class Question(JSONModel):
    """Describes a question on a StackExchange site."""
    transfer = ('tags', 'favorite_count', 'up_vote_count', 'down_vote_count',
        'view_count', 'score', 'community_owned', 'title', 'body',
        'body_markdown', 'is_answered', 'link', 'answer_count', 'can_close',
        'can_flag', 'close_vote_count', 'closed_reason', 'comment_count',
        'community_owned_date', 'delete_vote_count', 'down_vote_count',
        'downvoted', 'favorite_count', 'favorited', 'is_answered', 
        'accepted_answer_id', 'question_id', 'bounty_amount', 'upvoted',
        'reopen_vote_count', 'share_link', 'up_vote_count',
        ('creation_date', UNIXTimestamp),
        ('last_edit_date', UNIXTimestamp),
        ('last_activity_date', UNIXTimestamp),
        ('bounty_closes_date', UNIXTimestamp),
        ('locked_date', UNIXTimestamp),
        ('protected_date', UNIXTimestamp),
        # XXX
        #('bounty_user', PartialModelRef(User, lambda s: s.site.user(s.id), extend = True)),
        #('last_editor', PartialModelRef(User, lambda s: s.site.user(s.id), extend = True)),
        ('timeline', LazySequenceField(TimelineEvent, 'questions/{id}/timeline')),
        ('revisions', LazySequenceField(PostRevision, 'posts/{id}/revisions')),
        ('comments', LazySequenceField(Comment, 'questions/{id}/comments', filter = '!-*7AsUyrEan0')),
        ('answers', ListOf(ModelRef(Answer))))
    alias = (('id', 'question_id'),
             ('accepted_answer', 'accepted_answer_id', IDPartial(Answer, lambda a: a.site.answer(a.id))))

    def _extend(self, json, site):
        if hasattr(json, 'owner') and 'user_id' in json.owner:
            self.owner_id = json.owner['user_id']

            owner_dict = dict(json.owner)
            owner_dict['id'] = self.owner_id
            del owner_dict['user_id']
            owner_dict['user_type'] = UserType.from_string(owner_dict['user_type'])

            self.owner = User.partial(lambda self: self.site.user(self.id), site, owner_dict)

        self.url = 'http://' + self.site.root_domain + '/questions/' + str(self.id)

    def fetch_callback(self, _):
        return self.site.question(self.id)

    def linked(self):
        return self.site.questions(linked_to = self.id)

    def related(self):
        return self.site.questions(related_to = self.id)

    def __repr__(self):
        return "<Question '%s' @ %x>" % (self.title, id(self))

##### Tags #####
class TagSynonym(JSONModel):
    transfer = ('from_tag', 'to_tag', 'applied_count',
        ('creation_date', UNIXTimestamp),
        ('last_applied_date', UNIXTimestamp))

    def __repr__(self):
        return "<TagSynonym '%s'->'%s'>" % (self.from_tag, self.to_tag)

class TagWiki(JSONModel):
    transfer = ('tag_name', 'body', 'excerpt',
        ('body_last_edit_date', UNIXTimestamp),
        ('excerpt_last_edit_date', UNIXTimestamp))

    def _extend(self, json, site):
        if hasattr(json, 'last_body_editor'):
            body_editor = dict(json.last_body_editor)
            body_editor['id'] = body_editor['user_id']
            del body_editor['user_id']
            self.last_body_editor = User.partial(lambda s: s.site.user(self.id), site, body_editor)

        if hasattr(json, 'last_excerpt_editor'):
            excerpt_editor = dict(json.last_excerpt_editor)
            excerpt_editor['id'] = excerpt_editor['user_id']
            del excerpt_editor['user_id']
            self.last_excerpt_editor = User.partial(lambda s: s.site.user(self.id), site, excerpt_editor)

class Period(Enumeration):
    AllTime, Month = 'all-time', 'month'

class TopUser(JSONModel):
    transfer = ('score', 'post_count')

    def _extend(self, json, site):
        user_dict = dict(json.user)
        user_dict['id'] = user_dict['user_id']
        del user_dict['user_id']
        self.user = User.partial(lambda self: self.site.user(self.id), site, user_dict)

    def __repr__(self):
        return "<TopUser '%s' (score %d)>" % (self.user.display_name, self.score)

class Tag(JSONModel):
    transfer = ('name', 'count', 'fulfills_required')
    # Hack so that Site.vectorise() works correctly
    id = property(lambda self: self.name)

    def _extend(self, json, site):
        self.synonyms = StackExchangeLazySequence(TagSynonym, None, site, 'tags/%s/synonyms' % self.name, self._up('synonyms'), 'tag_synonyms')
        self.wiki = StackExchangeLazyObject(TagWiki, site, 'tags/%s/wikis' % self.name, self._up('wiki'), 'tag_wikis')

    def top_askers(self, period, **kw):
        return self.site.build('tags/%s/top-askers/%s' % (self.name, period), TopUser, 'top_users', kw)

    def top_answerers(self, period, **kw):
        return self.site.build('tags/%s/top-answerers/%s' % (self.name, period), TopUser, 'top_users', kw)

class RepChange(JSONModel):
    """Describes an event which causes a change in reputation."""
    transfer = ('user_id', 'post_id', 'post_type', 'title', 'positive_rep',
        'negative_rep', ('on_date', UNIXTimestamp))

    def _extend(self, json, site):
        if hasattr(json, 'positive_rep') and hasattr(json, 'negative_rep'):
            self.score = json.positive_rep - json.negative_rep

class UserType(Enumeration):
    """Denotes the status of a user on a site: whether it is Anonymous, Unregistered, Registered or a Moderator."""
    Anonymous = 'anonymous'
    Registered = 'registered'
    Unregistered = 'unregistered'
    Moderator = 'moderator'

class FormattedReputation(int):
    def format(rep):
        """Formats the reputation score like it is formatted on the sites. Heavily based on CMS' JavaScript implementation at
        http://stackapps.com/questions/1012/how-to-format-reputation-numbers-similar-to-stack-exchange-sites/1019#1019"""
        str_rep = str(rep)

        if rep < 1000:
            return str_rep
        elif rep < 10000:
            return '%s,%s' % (str_rep[0], str_rep[1:])
        elif rep % 1000 == 0:
            return '%dk' % (rep / 1000.0)
        else:
            return '%.1fk' % (rep / 1000.0)

class TopTag(JSONModel):
    transfer = ('tag_name', 'question_score', 'question_count', 'answer_score', 'answer_count')

    def __repr__(self):
        return "<TopTag '%s' Q:%d A:%d>" % (self.tag_name, self.question_score, self.answer_score)

class User(JSONModel):
    """Describes a user on a StackExchange site."""
    transfer = ('display_name', 'profile_image', 'age', 'website_url',
        'location', 'about_me', 'view_count', 'up_vote_count',
        'down_vote_count', 'account_id', 'profile_image',
        ('creation_date', UNIXTimestamp),
        ('last_access_date', UNIXTimestamp),
        ('reputation', FormattedReputation),
        ('favorites', LazySequenceField(Question, 'users/{id}/favorites', response_key = 'questions')),
        ('no_answers_questions', LazySequenceField(Question, 'users/{id}/questions/no-answers', response_key = 'questions')),
        ('unanswered_questions', LazySequenceField(Question, 'users/{id}/questions/unanswered', response_key = 'questions')),
        ('unaccepted_questions', LazySequenceField(Question, 'users/{id}/questions/unaccepted', response_key = 'questions')),
        ('tags', LazySequenceField(Tag, 'users/{id}/tags')),
        ('badges', LazySequenceField(LaterClass('Badge'), 'users/{id}/badges')),
        ('timeline', LazySequenceField(TimelineEvent, 'users/{id}/timeline', response_key = 'user_timelines')),
        ('reputation_detail', LazySequenceField(RepChange, 'users/{id}/reputation')),
        ('mentioned', LazySequenceField(Comment, 'users/{id}/mentioned', response_key = 'comments')),
        ('comments', LazySequenceField(Comment, 'users/{id}/comments')),
        ('top_answer_tags', LazySequenceField(TopTag, 'users/{id}/top-answer-tags', response_key = 'top_tags')),
        ('top_question_tags', LazySequenceField(TopTag, 'users/{id}/top-question-tags', response_key = 'top_tags')),
        )

    # for compatibility reasons; association_id changed in v2.x
    alias = (('id', 'user_id'), ('association_id', 'account_id'),
             ('type', 'user_type', UserType.from_string))
    badge_types = ('gold', 'silver', 'bronze')

    def _extend(self, json, site):
        user_questions_url = 'users/%d/questions' % self.id
        question_count = getattr(json, 'question_count', None)
        self.questions = StackExchangeLazySequence(Question, question_count, site, user_questions_url, self._up('questions'))

        user_answers_url = 'users/%d/answers' % self.id
        answer_count = getattr(json, 'answer_count', None)
        self.answers = StackExchangeLazySequence(Answer, answer_count, site, user_answers_url, self._up('answers'))

        if hasattr(self, 'up_vote_count') and hasattr(self, 'down_vote_count'):
            self.vote_counts = (self.up_vote_count, self.down_vote_count)

        if hasattr(json, 'badge_counts'):
            self.badge_counts_t = tuple(json.badge_counts.get(c, 0) for c in ('gold', 'silver', 'bronze'))
            self.gold_badges, self.silver_badges, self.bronze_badges = self.badge_counts_t
            self.badge_counts = {
                BadgeType.Gold:   self.gold_badges,
                BadgeType.Silver: self.silver_badges,
                BadgeType.Bronze: self.bronze_badges
            }
            self.badge_total = sum(self.badge_counts_t)

        if hasattr(self, 'type'):
            self.is_moderator = self.type == UserType.Moderator

        self.url = 'http://' + self.site.root_domain + '/users/' + str(self.id)

    def has_privilege(self, privilege):
        return self.reputation >= privilege.reputation

    def _get_real_tag(self, tag):
        return tag.name if isinstance(tag, Tag) else tag

    def top_answers_in_tag(self, tag, **kw):
        return self.site.build('users/%d/tags/%s/top-answers' % (self.id, self._get_real_tag(tag)), Answer, 'answers', kw)

    def top_questions_in_tag(self, tag, **kw):
        return self.site.build('users/%d/tags/%s/top-questions' % (self.id, self._get_real_tag(tag)), Question, 'questions', kw)

    def comments_to(self, user, **kw):
        uid = user.id if isinstance(user, User) else user
        return self.site.build('users/%d/comments/%d' % (self.id, uid), Comment, 'comments' ,kw)

    def __unicode__(self):
        return 'User %d [%s]' % (self.id, self.display_name)
    def __str__(self):
        return str(unicode(self))
    def __repr__(self):
        return "<User '%s' (%d) @ %x>" % (self.display_name, self.id, id(self))

class BadgeType(Enumeration):
    """Describes the rank or type of a badge: one of Bronze, Silver or Gold."""
    Bronze, Silver, Gold = range(3)

class Badge(JSONModel):
    """Describes a badge awardable on a StackExchange site."""
    transfer = ('name', 'description', 'award_count', 'tag_based',
                ('user', PartialModelRef(User, lambda s: s.site.user(s.id), extend = True)))
    alias = (('id', 'badge_id'),)

    @property
    def recipients(self):
        for badge in self.site.badge_recipients([self.id]):
            yield badge.user

    def __str__(self):
        return self.name
    def __repr__(self):
        return '<Badge \'%s\' @ %x>' % (self.name, id(self))

class Privilege(JSONModel):
    transfer = ('short_description', 'description', 'reputation')


class QuestionsQuery(object):
    def __init__(self, site):
        self.site = site

    def __call__(self, ids = None, user_id = None, **kw):
        self.site.check_filter(kw)

        # Compatibility hack, as user_id= was in versions below v1.1
        if ids is None and user_id is not None:
            return self.by_user(user_id, **kw)
        elif ids is None and user_id is None:
            return self.site.build('questions', Question, 'questions', kw)
        else:
            return self.site._get(Question, ids, 'questions', kw)

    def linked_to(self, qn, **kw):
        self.site.check_filter(kw)
        url = 'questions/%s/linked' % self.site.vectorise(qn, Question)
        return self.site.build(url, Question, 'questions', kw)

    def related_to(self, qn, **kw):
        self.site.check_filter(kw)
        url = 'questions/%s/related' % self.site.vectorise(qn, Question)
        return self.site.build(url, Question, 'questions', kw)

    def by_user(self, usr, **kw):
        self.site.check_filter(kw)
        kw['user_id'] = usr
        return self.site._user_prop('questions', Question, 'questions', kw)

    def unanswered(self, by = None, **kw):
        self.site.check_filter(kw)

        if by is None:
            return self.site.build('questions/unanswered', Question, 'questions', kw)
        else:
            kw['user_id'] = by
            return self.site._user_prop('questions/unanswered', Question, 'questions', kw)

    def no_answers(self, by = None, **kw):
        self.site.check_filter(kw)

        if by is None:
            return self.site.build('questions/no-answers', Question, 'questions', kw)
        else:
            kw['user_id'] = by
            return self.site._user_prop('questions/no-answers', Question, 'questions', kw)

    def unaccepted(self, by, **kw):
        self.site.check_filter(kw)
        kw['user_id'] = by
        return self.site._user_prop('questions/unaccepted', Questions, 'questions', kw)

    def favorited_by(self, by, **kw):
        self.site.check_filter(kw)
        kw['user_id'] = by
        return self.site._user_prop('favorites', Question, 'questions', kw)




import stackexchange
class __SEAPI(str):
    def __call__(self):
        return stackexchange.Site(self)
StackOverflow = __SEAPI('stackoverflow.com')
ServerFault = __SEAPI('serverfault.com')
SuperUser = __SEAPI('superuser.com')
MetaStackExchange = __SEAPI('meta.stackexchange.com')
WebApplications = __SEAPI('webapps.stackexchange.com')
WebApplicationsMeta = __SEAPI('meta.webapps.stackexchange.com')
Arqade = __SEAPI('gaming.stackexchange.com')
ArqadeMeta = __SEAPI('meta.gaming.stackexchange.com')
Webmasters = __SEAPI('webmasters.stackexchange.com')
WebmastersMeta = __SEAPI('meta.webmasters.stackexchange.com')
SeasonedAdvice = __SEAPI('cooking.stackexchange.com')
SeasonedAdviceMeta = __SEAPI('meta.cooking.stackexchange.com')
GameDevelopment = __SEAPI('gamedev.stackexchange.com')
GameDevelopmentMeta = __SEAPI('meta.gamedev.stackexchange.com')
Photography = __SEAPI('photo.stackexchange.com')
PhotographyMeta = __SEAPI('meta.photo.stackexchange.com')
CrossValidated = __SEAPI('stats.stackexchange.com')
CrossValidatedMeta = __SEAPI('meta.stats.stackexchange.com')
Mathematics = __SEAPI('math.stackexchange.com')
MathematicsMeta = __SEAPI('meta.math.stackexchange.com')
HomeImprovement = __SEAPI('diy.stackexchange.com')
HomeImprovementMeta = __SEAPI('meta.diy.stackexchange.com')
MetaSuperUser = __SEAPI('meta.superuser.com')
MetaServerFault = __SEAPI('meta.serverfault.com')
GeographicInformationSystems = __SEAPI('gis.stackexchange.com')
GeographicInformationSystemsMeta = __SEAPI('meta.gis.stackexchange.com')
TeXLaTeX = __SEAPI('tex.stackexchange.com')
TeXLaTeXMeta = __SEAPI('meta.tex.stackexchange.com')
AskUbuntu = __SEAPI('askubuntu.com')
AskUbuntuMeta = __SEAPI('meta.askubuntu.com')
PersonalFinanceampMoney = __SEAPI('money.stackexchange.com')
PersonalFinanceampMoneyMeta = __SEAPI('meta.money.stackexchange.com')
EnglishLanguageampUsage = __SEAPI('english.stackexchange.com')
EnglishLanguageampUsageMeta = __SEAPI('meta.english.stackexchange.com')
StackApps = __SEAPI('stackapps.com')
UserExperience = __SEAPI('ux.stackexchange.com')
UserExperienceMeta = __SEAPI('meta.ux.stackexchange.com')
UnixampLinux = __SEAPI('unix.stackexchange.com')
UnixampLinuxMeta = __SEAPI('meta.unix.stackexchange.com')
WordPressDevelopment = __SEAPI('wordpress.stackexchange.com')
WordPressDevelopmentMeta = __SEAPI('meta.wordpress.stackexchange.com')
TheoreticalComputerScience = __SEAPI('cstheory.stackexchange.com')
TheoreticalComputerScienceMeta = __SEAPI('meta.cstheory.stackexchange.com')
AskDifferent = __SEAPI('apple.stackexchange.com')
AskDifferentMeta = __SEAPI('meta.apple.stackexchange.com')
RoleplayingGames = __SEAPI('rpg.stackexchange.com')
RoleplayingGamesMeta = __SEAPI('meta.rpg.stackexchange.com')
Bicycles = __SEAPI('bicycles.stackexchange.com')
BicyclesMeta = __SEAPI('meta.bicycles.stackexchange.com')
Programmers = __SEAPI('programmers.stackexchange.com')
ProgrammersMeta = __SEAPI('meta.programmers.stackexchange.com')
ElectricalEngineering = __SEAPI('electronics.stackexchange.com')
ElectricalEngineeringMeta = __SEAPI('meta.electronics.stackexchange.com')
AndroidEnthusiasts = __SEAPI('android.stackexchange.com')
AndroidEnthusiastsMeta = __SEAPI('meta.android.stackexchange.com')
BoardampCardGames = __SEAPI('boardgames.stackexchange.com')
BoardampCardGamesMeta = __SEAPI('meta.boardgames.stackexchange.com')
Physics = __SEAPI('physics.stackexchange.com')
PhysicsMeta = __SEAPI('meta.physics.stackexchange.com')
Homebrewing = __SEAPI('homebrew.stackexchange.com')
HomebrewingMeta = __SEAPI('meta.homebrew.stackexchange.com')
InformationSecurity = __SEAPI('security.stackexchange.com')
InformationSecurityMeta = __SEAPI('meta.security.stackexchange.com')
Writers = __SEAPI('writers.stackexchange.com')
WritersMeta = __SEAPI('meta.writers.stackexchange.com')
VideoProduction = __SEAPI('video.stackexchange.com')
VideoProductionMeta = __SEAPI('meta.video.stackexchange.com')
GraphicDesign = __SEAPI('graphicdesign.stackexchange.com')
GraphicDesignMeta = __SEAPI('meta.graphicdesign.stackexchange.com')
DatabaseAdministrators = __SEAPI('dba.stackexchange.com')
DatabaseAdministratorsMeta = __SEAPI('meta.dba.stackexchange.com')
ScienceFictionampFantasy = __SEAPI('scifi.stackexchange.com')
ScienceFictionampFantasyMeta = __SEAPI('meta.scifi.stackexchange.com')
CodeReview = __SEAPI('codereview.stackexchange.com')
CodeReviewMeta = __SEAPI('meta.codereview.stackexchange.com')
ProgrammingPuzzlesampCodeGolf = __SEAPI('codegolf.stackexchange.com')
ProgrammingPuzzlesampCodeGolfMeta = __SEAPI('meta.codegolf.stackexchange.com')
QuantitativeFinance = __SEAPI('quant.stackexchange.com')
QuantitativeFinanceMeta = __SEAPI('meta.quant.stackexchange.com')
ProjectManagement = __SEAPI('pm.stackexchange.com')
ProjectManagementMeta = __SEAPI('meta.pm.stackexchange.com')
Skeptics = __SEAPI('skeptics.stackexchange.com')
SkepticsMeta = __SEAPI('meta.skeptics.stackexchange.com')
PhysicalFitness = __SEAPI('fitness.stackexchange.com')
PhysicalFitnessMeta = __SEAPI('meta.fitness.stackexchange.com')
DrupalAnswers = __SEAPI('drupal.stackexchange.com')
DrupalAnswersMeta = __SEAPI('meta.drupal.stackexchange.com')
MotorVehicleMaintenanceampRepair = __SEAPI('mechanics.stackexchange.com')
MotorVehicleMaintenanceampRepairMeta = __SEAPI('meta.mechanics.stackexchange.com')
Parenting = __SEAPI('parenting.stackexchange.com')
ParentingMeta = __SEAPI('meta.parenting.stackexchange.com')
SharePoint = __SEAPI('sharepoint.stackexchange.com')
SharePointMeta = __SEAPI('meta.sharepoint.stackexchange.com')
MusicPracticeampTheory = __SEAPI('music.stackexchange.com')
MusicPracticeampTheoryMeta = __SEAPI('meta.music.stackexchange.com')
SoftwareQualityAssuranceampTesting = __SEAPI('sqa.stackexchange.com')
SoftwareQualityAssuranceampTestingMeta = __SEAPI('meta.sqa.stackexchange.com')
MiYodeya = __SEAPI('judaism.stackexchange.com')
MiYodeyaMeta = __SEAPI('meta.judaism.stackexchange.com')
GermanLanguage = __SEAPI('german.stackexchange.com')
GermanLanguageMeta = __SEAPI('meta.german.stackexchange.com')
JapaneseLanguage = __SEAPI('japanese.stackexchange.com')
JapaneseLanguageMeta = __SEAPI('meta.japanese.stackexchange.com')
Philosophy = __SEAPI('philosophy.stackexchange.com')
PhilosophyMeta = __SEAPI('meta.philosophy.stackexchange.com')
GardeningampLandscaping = __SEAPI('gardening.stackexchange.com')
GardeningampLandscapingMeta = __SEAPI('meta.gardening.stackexchange.com')
Travel = __SEAPI('travel.stackexchange.com')
TravelMeta = __SEAPI('meta.travel.stackexchange.com')
PersonalProductivity = __SEAPI('productivity.stackexchange.com')
PersonalProductivityMeta = __SEAPI('meta.productivity.stackexchange.com')
Cryptography = __SEAPI('crypto.stackexchange.com')
CryptographyMeta = __SEAPI('meta.crypto.stackexchange.com')
SignalProcessing = __SEAPI('dsp.stackexchange.com')
SignalProcessingMeta = __SEAPI('meta.dsp.stackexchange.com')
FrenchLanguage = __SEAPI('french.stackexchange.com')
FrenchLanguageMeta = __SEAPI('meta.french.stackexchange.com')
Christianity = __SEAPI('christianity.stackexchange.com')
ChristianityMeta = __SEAPI('meta.christianity.stackexchange.com')
Bitcoin = __SEAPI('bitcoin.stackexchange.com')
StackOverflow = __SEAPI('stackoverflow.com')
ServerFault = __SEAPI('serverfault.com')
SuperUser = __SEAPI('superuser.com')
MetaStackExchange = __SEAPI('meta.stackexchange.com')
WebApplications = __SEAPI('webapps.stackexchange.com')
WebApplicationsMeta = __SEAPI('meta.webapps.stackexchange.com')
Arqade = __SEAPI('gaming.stackexchange.com')
ArqadeMeta = __SEAPI('meta.gaming.stackexchange.com')
Webmasters = __SEAPI('webmasters.stackexchange.com')
WebmastersMeta = __SEAPI('meta.webmasters.stackexchange.com')
SeasonedAdvice = __SEAPI('cooking.stackexchange.com')
SeasonedAdviceMeta = __SEAPI('meta.cooking.stackexchange.com')
GameDevelopment = __SEAPI('gamedev.stackexchange.com')
GameDevelopmentMeta = __SEAPI('meta.gamedev.stackexchange.com')
Photography = __SEAPI('photo.stackexchange.com')
PhotographyMeta = __SEAPI('meta.photo.stackexchange.com')
CrossValidated = __SEAPI('stats.stackexchange.com')
CrossValidatedMeta = __SEAPI('meta.stats.stackexchange.com')
Mathematics = __SEAPI('math.stackexchange.com')
MathematicsMeta = __SEAPI('meta.math.stackexchange.com')
HomeImprovement = __SEAPI('diy.stackexchange.com')
HomeImprovementMeta = __SEAPI('meta.diy.stackexchange.com')
MetaSuperUser = __SEAPI('meta.superuser.com')
MetaServerFault = __SEAPI('meta.serverfault.com')
GeographicInformationSystems = __SEAPI('gis.stackexchange.com')
GeographicInformationSystemsMeta = __SEAPI('meta.gis.stackexchange.com')
TeXLaTeX = __SEAPI('tex.stackexchange.com')
TeXLaTeXMeta = __SEAPI('meta.tex.stackexchange.com')
AskUbuntu = __SEAPI('askubuntu.com')
AskUbuntuMeta = __SEAPI('meta.askubuntu.com')
PersonalFinanceampMoney = __SEAPI('money.stackexchange.com')
PersonalFinanceampMoneyMeta = __SEAPI('meta.money.stackexchange.com')
EnglishLanguageampUsage = __SEAPI('english.stackexchange.com')
EnglishLanguageampUsageMeta = __SEAPI('meta.english.stackexchange.com')
StackApps = __SEAPI('stackapps.com')
UserExperience = __SEAPI('ux.stackexchange.com')
UserExperienceMeta = __SEAPI('meta.ux.stackexchange.com')
UnixampLinux = __SEAPI('unix.stackexchange.com')
UnixampLinuxMeta = __SEAPI('meta.unix.stackexchange.com')
WordPressDevelopment = __SEAPI('wordpress.stackexchange.com')
WordPressDevelopmentMeta = __SEAPI('meta.wordpress.stackexchange.com')
TheoreticalComputerScience = __SEAPI('cstheory.stackexchange.com')
TheoreticalComputerScienceMeta = __SEAPI('meta.cstheory.stackexchange.com')
AskDifferent = __SEAPI('apple.stackexchange.com')
AskDifferentMeta = __SEAPI('meta.apple.stackexchange.com')
RoleplayingGames = __SEAPI('rpg.stackexchange.com')
RoleplayingGamesMeta = __SEAPI('meta.rpg.stackexchange.com')
Bicycles = __SEAPI('bicycles.stackexchange.com')
BicyclesMeta = __SEAPI('meta.bicycles.stackexchange.com')
Programmers = __SEAPI('programmers.stackexchange.com')
ProgrammersMeta = __SEAPI('meta.programmers.stackexchange.com')
ElectricalEngineering = __SEAPI('electronics.stackexchange.com')
ElectricalEngineeringMeta = __SEAPI('meta.electronics.stackexchange.com')
AndroidEnthusiasts = __SEAPI('android.stackexchange.com')
AndroidEnthusiastsMeta = __SEAPI('meta.android.stackexchange.com')
BoardampCardGames = __SEAPI('boardgames.stackexchange.com')
BoardampCardGamesMeta = __SEAPI('meta.boardgames.stackexchange.com')
Physics = __SEAPI('physics.stackexchange.com')
PhysicsMeta = __SEAPI('meta.physics.stackexchange.com')
Homebrewing = __SEAPI('homebrew.stackexchange.com')
HomebrewingMeta = __SEAPI('meta.homebrew.stackexchange.com')
InformationSecurity = __SEAPI('security.stackexchange.com')
InformationSecurityMeta = __SEAPI('meta.security.stackexchange.com')
Writers = __SEAPI('writers.stackexchange.com')
WritersMeta = __SEAPI('meta.writers.stackexchange.com')
VideoProduction = __SEAPI('video.stackexchange.com')
VideoProductionMeta = __SEAPI('meta.video.stackexchange.com')
GraphicDesign = __SEAPI('graphicdesign.stackexchange.com')
GraphicDesignMeta = __SEAPI('meta.graphicdesign.stackexchange.com')
DatabaseAdministrators = __SEAPI('dba.stackexchange.com')
DatabaseAdministratorsMeta = __SEAPI('meta.dba.stackexchange.com')
ScienceFictionampFantasy = __SEAPI('scifi.stackexchange.com')
ScienceFictionampFantasyMeta = __SEAPI('meta.scifi.stackexchange.com')
CodeReview = __SEAPI('codereview.stackexchange.com')
CodeReviewMeta = __SEAPI('meta.codereview.stackexchange.com')
ProgrammingPuzzlesampCodeGolf = __SEAPI('codegolf.stackexchange.com')
ProgrammingPuzzlesampCodeGolfMeta = __SEAPI('meta.codegolf.stackexchange.com')
QuantitativeFinance = __SEAPI('quant.stackexchange.com')
QuantitativeFinanceMeta = __SEAPI('meta.quant.stackexchange.com')
ProjectManagement = __SEAPI('pm.stackexchange.com')
ProjectManagementMeta = __SEAPI('meta.pm.stackexchange.com')
Skeptics = __SEAPI('skeptics.stackexchange.com')
SkepticsMeta = __SEAPI('meta.skeptics.stackexchange.com')
PhysicalFitness = __SEAPI('fitness.stackexchange.com')
PhysicalFitnessMeta = __SEAPI('meta.fitness.stackexchange.com')
DrupalAnswers = __SEAPI('drupal.stackexchange.com')
DrupalAnswersMeta = __SEAPI('meta.drupal.stackexchange.com')
MotorVehicleMaintenanceampRepair = __SEAPI('mechanics.stackexchange.com')
MotorVehicleMaintenanceampRepairMeta = __SEAPI('meta.mechanics.stackexchange.com')
Parenting = __SEAPI('parenting.stackexchange.com')
ParentingMeta = __SEAPI('meta.parenting.stackexchange.com')
SharePoint = __SEAPI('sharepoint.stackexchange.com')
SharePointMeta = __SEAPI('meta.sharepoint.stackexchange.com')
MusicPracticeampTheory = __SEAPI('music.stackexchange.com')
MusicPracticeampTheoryMeta = __SEAPI('meta.music.stackexchange.com')
SoftwareQualityAssuranceampTesting = __SEAPI('sqa.stackexchange.com')
SoftwareQualityAssuranceampTestingMeta = __SEAPI('meta.sqa.stackexchange.com')
MiYodeya = __SEAPI('judaism.stackexchange.com')
MiYodeyaMeta = __SEAPI('meta.judaism.stackexchange.com')
GermanLanguage = __SEAPI('german.stackexchange.com')
GermanLanguageMeta = __SEAPI('meta.german.stackexchange.com')
JapaneseLanguage = __SEAPI('japanese.stackexchange.com')
JapaneseLanguageMeta = __SEAPI('meta.japanese.stackexchange.com')
Philosophy = __SEAPI('philosophy.stackexchange.com')
PhilosophyMeta = __SEAPI('meta.philosophy.stackexchange.com')
GardeningampLandscaping = __SEAPI('gardening.stackexchange.com')
GardeningampLandscapingMeta = __SEAPI('meta.gardening.stackexchange.com')
Travel = __SEAPI('travel.stackexchange.com')
TravelMeta = __SEAPI('meta.travel.stackexchange.com')
PersonalProductivity = __SEAPI('productivity.stackexchange.com')
PersonalProductivityMeta = __SEAPI('meta.productivity.stackexchange.com')
Cryptography = __SEAPI('crypto.stackexchange.com')
CryptographyMeta = __SEAPI('meta.crypto.stackexchange.com')
SignalProcessing = __SEAPI('dsp.stackexchange.com')
SignalProcessingMeta = __SEAPI('meta.dsp.stackexchange.com')
FrenchLanguage = __SEAPI('french.stackexchange.com')
FrenchLanguageMeta = __SEAPI('meta.french.stackexchange.com')
Christianity = __SEAPI('christianity.stackexchange.com')
ChristianityMeta = __SEAPI('meta.christianity.stackexchange.com')
Bitcoin = __SEAPI('bitcoin.stackexchange.com')
BitcoinMeta = __SEAPI('meta.bitcoin.stackexchange.com')
Linguistics = __SEAPI('linguistics.stackexchange.com')
LinguisticsMeta = __SEAPI('meta.linguistics.stackexchange.com')
BiblicalHermeneutics = __SEAPI('hermeneutics.stackexchange.com')
BiblicalHermeneuticsMeta = __SEAPI('meta.hermeneutics.stackexchange.com')
History = __SEAPI('history.stackexchange.com')
HistoryMeta = __SEAPI('meta.history.stackexchange.com')
LEGO174Answers = __SEAPI('bricks.stackexchange.com')
LEGO174AnswersMeta = __SEAPI('meta.bricks.stackexchange.com')
SpanishLanguage = __SEAPI('spanish.stackexchange.com')
SpanishLanguageMeta = __SEAPI('meta.spanish.stackexchange.com')
ComputationalScience = __SEAPI('scicomp.stackexchange.com')
ComputationalScienceMeta = __SEAPI('meta.scicomp.stackexchange.com')
MoviesampTV = __SEAPI('movies.stackexchange.com')
MoviesampTVMeta = __SEAPI('meta.movies.stackexchange.com')
ChineseLanguage = __SEAPI('chinese.stackexchange.com')
ChineseLanguageMeta = __SEAPI('meta.chinese.stackexchange.com')
Biology = __SEAPI('biology.stackexchange.com')
BiologyMeta = __SEAPI('meta.biology.stackexchange.com')
Poker = __SEAPI('poker.stackexchange.com')
PokerMeta = __SEAPI('meta.poker.stackexchange.com')
Mathematica = __SEAPI('mathematica.stackexchange.com')
MathematicaMeta = __SEAPI('meta.mathematica.stackexchange.com')
CognitiveSciences = __SEAPI('cogsci.stackexchange.com')
CognitiveSciencesMeta = __SEAPI('meta.cogsci.stackexchange.com')
TheGreatOutdoors = __SEAPI('outdoors.stackexchange.com')
TheGreatOutdoorsMeta = __SEAPI('meta.outdoors.stackexchange.com')
MartialArts = __SEAPI('martialarts.stackexchange.com')
MartialArtsMeta = __SEAPI('meta.martialarts.stackexchange.com')
Sports = __SEAPI('sports.stackexchange.com')
SportsMeta = __SEAPI('meta.sports.stackexchange.com')
Academia = __SEAPI('academia.stackexchange.com')
AcademiaMeta = __SEAPI('meta.academia.stackexchange.com')
ComputerScience = __SEAPI('cs.stackexchange.com')
ComputerScienceMeta = __SEAPI('meta.cs.stackexchange.com')
TheWorkplace = __SEAPI('workplace.stackexchange.com')
TheWorkplaceMeta = __SEAPI('meta.workplace.stackexchange.com')
WindowsPhone = __SEAPI('windowsphone.stackexchange.com')
WindowsPhoneMeta = __SEAPI('meta.windowsphone.stackexchange.com')
Chemistry = __SEAPI('chemistry.stackexchange.com')
ChemistryMeta = __SEAPI('meta.chemistry.stackexchange.com')
Chess = __SEAPI('chess.stackexchange.com')
ChessMeta = __SEAPI('meta.chess.stackexchange.com')
RaspberryPi = __SEAPI('raspberrypi.stackexchange.com')
RaspberryPiMeta = __SEAPI('meta.raspberrypi.stackexchange.com')
RussianLanguage = __SEAPI('russian.stackexchange.com')
RussianLanguageMeta = __SEAPI('meta.russian.stackexchange.com')
Islam = __SEAPI('islam.stackexchange.com')
IslamMeta = __SEAPI('meta.islam.stackexchange.com')
Salesforce = __SEAPI('salesforce.stackexchange.com')
SalesforceMeta = __SEAPI('meta.salesforce.stackexchange.com')
AskPatents = __SEAPI('patents.stackexchange.com')
AskPatentsMeta = __SEAPI('meta.patents.stackexchange.com')
GenealogyampFamilyHistory = __SEAPI('genealogy.stackexchange.com')
GenealogyampFamilyHistoryMeta = __SEAPI('meta.genealogy.stackexchange.com')
Robotics = __SEAPI('robotics.stackexchange.com')
RoboticsMeta = __SEAPI('meta.robotics.stackexchange.com')
ExpressionEngine174Answers = __SEAPI('expressionengine.stackexchange.com')
ExpressionEngine174AnswersMeta = __SEAPI('meta.expressionengine.stackexchange.com')
Politics = __SEAPI('politics.stackexchange.com')
PoliticsMeta = __SEAPI('meta.politics.stackexchange.com')
AnimeampManga = __SEAPI('anime.stackexchange.com')
AnimeampMangaMeta = __SEAPI('meta.anime.stackexchange.com')
Magento = __SEAPI('magento.stackexchange.com')
MagentoMeta = __SEAPI('meta.magento.stackexchange.com')
EnglishLanguageLearners = __SEAPI('ell.stackexchange.com')
EnglishLanguageLearnersMeta = __SEAPI('meta.ell.stackexchange.com')
SustainableLiving = __SEAPI('sustainability.stackexchange.com')
SustainableLivingMeta = __SEAPI('meta.sustainability.stackexchange.com')
Tridion = __SEAPI('tridion.stackexchange.com')
TridionMeta = __SEAPI('meta.tridion.stackexchange.com')
ReverseEngineering = __SEAPI('reverseengineering.stackexchange.com')
ReverseEngineeringMeta = __SEAPI('meta.reverseengineering.stackexchange.com')
NetworkEngineering = __SEAPI('networkengineering.stackexchange.com')
NetworkEngineeringMeta = __SEAPI('meta.networkengineering.stackexchange.com')
OpenData = __SEAPI('opendata.stackexchange.com')
OpenDataMeta = __SEAPI('meta.opendata.stackexchange.com')
Freelancing = __SEAPI('freelancing.stackexchange.com')
FreelancingMeta = __SEAPI('meta.freelancing.stackexchange.com')
Blender = __SEAPI('blender.stackexchange.com')
BlenderMeta = __SEAPI('meta.blender.stackexchange.com')
MathOverflow = __SEAPI('mathoverflow.net')
MathOverflowMeta = __SEAPI('meta.mathoverflow.net')
SpaceExploration = __SEAPI('space.stackexchange.com')
SpaceExplorationMeta = __SEAPI('meta.space.stackexchange.com')
SoundDesign = __SEAPI('sound.stackexchange.com')
SoundDesignMeta = __SEAPI('meta.sound.stackexchange.com')
Astronomy = __SEAPI('astronomy.stackexchange.com')
AstronomyMeta = __SEAPI('meta.astronomy.stackexchange.com')
Tor = __SEAPI('tor.stackexchange.com')
TorMeta = __SEAPI('meta.tor.stackexchange.com')
Pets = __SEAPI('pets.stackexchange.com')
PetsMeta = __SEAPI('meta.pets.stackexchange.com')
AmateurRadio = __SEAPI('ham.stackexchange.com')
AmateurRadioMeta = __SEAPI('meta.ham.stackexchange.com')
ItalianLanguage = __SEAPI('italian.stackexchange.com')
ItalianLanguageMeta = __SEAPI('meta.italian.stackexchange.com')
StackOverflowemPortugu234s = __SEAPI('pt.stackoverflow.com')
StackOverflowemPortugu234sMeta = __SEAPI('meta.pt.stackoverflow.com')
Aviation = __SEAPI('aviation.stackexchange.com')
AviationMeta = __SEAPI('meta.aviation.stackexchange.com')
Ebooks = __SEAPI('ebooks.stackexchange.com')
EbooksMeta = __SEAPI('meta.ebooks.stackexchange.com')
Beer = __SEAPI('beer.stackexchange.com')
BeerMeta = __SEAPI('meta.beer.stackexchange.com')
SoftwareRecommendations = __SEAPI('softwarerecs.stackexchange.com')
SoftwareRecommendationsMeta = __SEAPI('meta.softwarerecs.stackexchange.com')
Arduino = __SEAPI('arduino.stackexchange.com')
ArduinoMeta = __SEAPI('meta.arduino.stackexchange.com')
CS50 = __SEAPI('cs50.stackexchange.com')
CS50Meta = __SEAPI('meta.cs50.stackexchange.com')
edxcs1691x = __SEAPI('edx-cs169-1x.stackexchange.com')
edxcs1691xMeta = __SEAPI('meta.edx-cs169-1x.stackexchange.com')
Expatriates = __SEAPI('expatriates.stackexchange.com')
ExpatriatesMeta = __SEAPI('meta.expatriates.stackexchange.com')
MathematicsEducators = __SEAPI('matheducators.stackexchange.com')
MathematicsEducatorsMeta = __SEAPI('meta.matheducators.stackexchange.com')
MetaStackOverflow = __SEAPI('meta.stackoverflow.com')
EarthScience = __SEAPI('earthscience.stackexchange.com')
EarthScienceMeta = __SEAPI('meta.earthscience.stackexchange.com')
